#ifndef MENU_HPP_INCLUDED
#define MENU_HPP_INCLUDED

#include "../../Core/Types/Box.hpp"
#include <cstdint>
#include <string>
#include <vector>

/**
* @brief A namespace containing various functions for the right-click menu
*/
namespace Menu
{

    bool IsOpen();

    /**
     * @brief Opens the menu by right-clicking wherever the mouse currently is
     * @return true if the menu is currently open or opened by right clicking
     * @return false if WaitFunc fails - WaitFunc(1000, 100, Menu::IsOpen, true);
     */
    bool Open();
    bool Close();



    std::int32_t GetCount();

    /**
     * @brief Attempts to verify the order of the passed menu actions and targets
     * @return true if the passed menu actions and targets are in the correct order
     * @note This isn't a guarantee, but testing shows near zero failure rate
     */
    bool OptionsValid(std::int32_t MenuCount, const std::vector<std::string>& RawActions, const std::vector<std::string>& RawTargets);

    /**
     * @brief Get the menu actions
     *
     * Actions are almost always the very first word in a menu option
     * 'Use' - action 'gold bar' - target
     * 'Attack' - action 'Guard  (level-21)' - target
     * @return std::vector<std::string> array of menu actions
     * @warning The result can rarely be inaccurate if the menu isn't open before calling this function
     */
    std::vector<std::string> GetActions();

    /**
     * @brief Get the menu targets
     *
     * Targets are almost always everything after the action
     * 'Use' - action 'gold bar' - target
     * 'Attack' - action 'Guard  (level-21)' - target
     * @return std::vector<std::string> array of menu targets
     * @warning The result can rarely be inaccurate if the menu isn't open before calling this function
     */
    std::vector<std::string> GetTargets();

    /**
     * @brief Get the menu options
     *
     * 'Use' - action 'gold bar' - target
     * 'Attack' - action 'Guard  (level-21)' - target
     * 'Use gold bar' - option
     * 'Attack Guard  (level-21)' - option
     * @return std::vector<std::string> array of menu options
     * @warning The result can rarely be inaccurate if the menu isn't open before calling this function
     */
    std::vector<std::string> GetOptions();
    /**
     * @brief Get the menu options split into pairs instead of whole strings
     *
     * 'Use' - action 'gold bar' - target
     * 'Attack' - action 'Guard  (level-21)' - target
     * 'Use gold bar' - option
     * 'Attack Guard  (level-21)' - option
     * first is the action, second is the target
     * @return std::vector<std::pair<std::string, std::string>>
     * @warning The result can rarely be inaccurate if the menu isn't open before calling this function
     */
    std::vector<std::pair<std::string, std::string>> GetSplitOptions();


    /**
     * @brief Get the index of the found option
     *
     * This function will find the first menu option containing the string 'Option'
     * This means passing 'Att' or 'Attack' will return the first found menu option containing 'Att' or 'Attack'
     * If you want to find a very specific menu option, the more the 'Option' string contains, the more specific it'll be
     *
     * Say for example, you want to attack a guard whose level is 40, and on this one tile, there's 5 different guards, all with different levels
     * If you pass 'Attack' as the 'Option' to look for, it'll select the very first guard, regardless of level.
     * If you pass 'Attack Guard  (level-40)' as the 'Option' to look for, it'll attack the guard whose level is 40.
     * @param Option %Menu option to look for, can be very specific, or very lenient
     * @return std::int32_t The index of the first found option, -1 if the option wasn't found
     * @note Notice that 'Attack Guard  (level-40)' has a double space after 'Guard', menu options can vary, in this case the menu option i'm looking for
     * has a double space after 'Guard'.
     * @warning The result can rarely be inaccurate if the menu isn't open before calling this function
     */
    std::int32_t IndexOf(const std::string& Option);

    /**
     * @brief Get the index of the first found option
     *
     * This function will return the first found index, of the first found option.
     * This function is just Menu::IndexOf(const std::string& Option), but instead of looking for one single option, this function looks for multiple
     *
     * @param Options %Menu options to look for, returns on first option found, the more specific the options, the more accurate the result will be
     * @return std::int32_t The index of the first found option, -1 if the option wasn't found
     * @see Menu::IndexOf(const std::string& Option)
     * @warning The result can rarely be inaccurate if the menu isn't open before calling this function
     */
    std::int32_t IndexOf(const std::vector<std::string>& Options);


    /**
     * @brief Looks for a menu option at the passed index, and returns information about the found option
     *
     * @param Index The index of the menu option to look at
     * @return std::tuple<bool, std::string, std::string>
     * std::int32_t [0] = Index of the option that was found, -1 if the option wasn't found
     * std::string [1] = Action that was found at index
     * std::string [2] = Target that was found at index
     * @see Menu::IndexOf(const std::string& Option)
     * @warning The result can rarely be inaccurate if the menu isn't open before calling this function
     */
    std::tuple<std::int32_t, std::string, std::string> FindOption(std::uint32_t Index);

    /**
     * @brief Looks for a menu option containing the passed Option, and returns information about the found option
     *
     * @param Option %Menu option to look for, the more specific, the more accurate the result will be
     * @return std::tuple<bool, std::string, std::string>
     * std::int32_t [0] = Index of the option that was found, -1 if the option wasn't found
     * std::string [1] = Action that was found
     * std::string [2] = Target that was found
     * @see Menu::IndexOf(const std::string& Option)
     * @warning The result can rarely be inaccurate if the menu isn't open before calling this function
     */
    std::tuple<std::int32_t, std::string, std::string> FindOption(const std::string& Option);

    /**
     * @brief Looks for a menu option containing the passed Option, and returns information about the found option
     *
     * @param %Options %Menu options to look for, returns on first option found, the more specific the options, the more accurate the result will be
     * @return std::tuple<bool, std::string, std::string>
     * std::int32_t [0] = Index of the option that was found, -1 if the option wasn't found
     * std::string [1] = Action that was found
     * std::string [2] = Target that was found
     * @see Menu::IndexOf(const std::string& Options)
     * @warning The result can rarely be inaccurate if the menu isn't open before calling this function
     */
    std::tuple<std::int32_t, std::string, std::string> FindOption(const std::vector<std::string>& Options);

    /**
     * @brief Looks for a menu option containing the passed Option, and returns information of all options that were found
     *
     * @param Option %Menu option to look for, the more specific the option, the more accurate the result will be
     * @return std::vector<std::tuple<std::int32_t, std::string, std::string>>
     * std::int32_t [0] = Index of the option that was found, -1 if the option wasn't found
     * std::string [1] = Action that was found
     * std::string [2] = Target that was found
     * @warning The result can rarely be inaccurate if the menu isn't open before calling this function
     * @see Menu::IndexOf(const std::string& Option)
     */
    std::vector<std::tuple<std::int32_t, std::string, std::string>> FindOptions(const std::string& Option);

    /**
     * @brief Looks for a menu option containing the passed Option, and returns information of all options that were found
     *
     * @param %Options %Menu options to look for, if the option is found within the menu options, it'll be added to the result array, the more specific the options, the more accurate the result will be
     * @return std::vector<std::tuple<std::int32_t, std::string, std::string>>
     * std::int32_t [0] = Index of the option that was found, -1 if the option wasn't found
     * std::string [1] = Action that was found
     * std::string [2] = Target that was found
     * @see Menu::IndexOf(const std::string& Options)
     * @warning The result can rarely be inaccurate if the menu isn't open before calling this function
     */
    std::vector<std::tuple<std::int32_t, std::string, std::string>> FindOptions(const std::vector<std::string>& Options);


    /**
     * @brief Returns true if the menu contains the passed Option
     *
     * @param Option %Menu option to look for, the more specific the option, the more specific the option, the more accurate the result will be
     * @return true if the found index is valid (Index >= 0)
     * @see Menu::IndexOf(const std::string& Option)
     * @warning The result can rarely be inaccurate if the menu isn't open before calling this function
     */
    bool Contains(const std::string& Option);

    /**
     * @brief Returns true if the menu contains at least one of the passed Options
     *
     * @param %Options %Menu options to look for, returns on first option found, the more specific the options, the more accurate the result will be
     * @return true if the found index is valid (Index >= 0)
     * @see Menu::IndexOf(const std::string& Options)
     * @warning The result can rarely be inaccurate if the menu isn't open before calling this function
     */
    bool Contains(const std::vector<std::string>& Options);


    /**
     * @brief Waits until the menu contains the passed option
     *
     * @param Duration How long to wait until returning
     * @param Step How long to wait until it checks the menu options
     * @param Option %Menu option to look for, the more specific the option, the more accurate the result will be
     * @return true if the found index is valid (Index >= 0)
     * @see Menu::IndexOf(const std::string& Option)
     * @see Menu::Contains(const std::string& Option)
     */
    bool WaitContains(std::uint32_t Duration, std::uint32_t Step, const std::string& Option);

    /**
     * @brief Waits until the menu contains at leaset one of the passed options
     *
     * @param Duration How long to wait until returning
     * @param Step How long to wait until it checks the menu options
     * @param %Options %Menu options to look for, returns on first option found, the more specific the options, the more accurate the result will be
     * @return true if the found index is valid (Index >= 0)
     * @see Menu::IndexOf(const std::string& Options)
     * @see Menu::Contains(const std::string& Options)
     */
    bool WaitContains(std::uint32_t Duration, std::uint32_t Step, const std::vector<std::string>& Options);


    /**
     * @brief Selects the menu option by the passed index if the menu is open
     *
     * @param Index The index of the option
     * @param CloseMenu Close the menu if the function fails by using Menu::Close(), true by default
     * @return true if the function selected the option by index
     * @return false if the menu isn't open, the passed index is invalid (-1), the option box is off-screen, or the WaitFunc failed - WaitFunc(1000, 100, Menu::IsOpen, false)
     */
    bool Select(std::int32_t Index, bool CloseMenu = true);

    /**
     * @brief Selects the passed menu option if the menu is open
     *
     * @param Option %Menu option to look for, the more specific the option, the more accurate the result will be
     * @param CloseMenu Close the menu if the function fails by using Menu::Close(), true by default
     * @return true if the function selected the option by option
     * @return false if the menu isn't open, the found option was invalid, the option box is off-screen, or the WaitFunc failed - WaitFunc(1000, 100, Menu::IsOpen, false)
     * @see Menu::IndexOf(const std::string& Option)
     */
    bool Select(const std::string& Option, bool CloseMenu = true);

    /**
     * @brief Selects the first found option in the menu
     *
     * @param %Options %Menu options to look for, uses the first option found, the more specific the options, the more accurate the result will be
     * @param CloseMenu Close the menu if the function fails using Menu::Close(), true by default
     * @return true if the function selected the option by the first found option
     * @return false if the menu isn't open, the found option was invalid, the option box is off-screen, or the WaitFunc failed - WaitFunc(1000, 100, Menu::IsOpen, false)
     * @see Menu::IndexOf(const std::string& Options)
     */
    bool Select(const std::vector<std::string>& Options, bool CloseMenu = true);


    /**
     * @brief Get the array of menu option boxes, each box is a menu option
     * @return std::vector<Box> array of the menu option boxes, returns an empty array if the menu isn't open
     */
    std::vector<Box> GetOptionBoxes();

    /**
     * @return Box representing the full menu, returns a null box if the menu isn't open
     */
    Box GetBox();
}

#endif // MENU_HPP_INCLUDED
